// $Id: SequenceInfo.h 973 2017-09-12 13:27:49Z ge $
/// \file SequenceInfo.h
/// \brief This class manages experimental sequence information
///
/// $Revision: 973 $
/// \author Gerald Weber <gweberbh@gmail.com>
#ifndef GBC_SEQUENCEINFO_H
#define GBC_SEQUENCEINFO_H "$Id: SequenceInfo.h 973 2017-09-12 13:27:49Z ge $"

#include "NucleotideSequence.h"
#include "NeighbourSequence.h"
#include "Duplex.h"
#include "StrandPairSequence.h"
#include "ReferenceSet.h"

using namespace gbc;
namespace gbc
{
template<class _Tp=double>
class Data
  {
  typedef _Tp value_type;
  public:
    std::string name;
    value_type measured;
    value_type adjusted;
    value_type predicted;

  Data(void): name(), measured(), adjusted(), predicted() {}

  Data(std::string nm): name(nm), measured(), adjusted(), predicted() {}

  inline void print_name(std::ostream &out) const
    {
    out << name << ".measured " 
        << name << ".adjusted "
        << name << ".predicted "; 
    }

  inline friend std::ostream& operator<<(std::ostream &out, Data& dt)
    {
    out << dt.measured << " " << dt.adjusted << " " << dt.predicted;
    return out;
    }
  };

template<bool _Periodic=false, class _Tp=double, class _sequence_type=std::pair<std::string,std::string> >
class SequenceInfo
  {
  public:
    typedef _Tp                 value_type;
    typedef Data<value_type>    data_type;
    typedef _sequence_type      sequence_type;
    typedef Duplex<>            duplex_type;
    typedef BasePair<>          base_pair_type;
    typedef std::deque<duplex_type> duplex_deque_type;
    typedef NeighbourSequence<> neighboursequence_type;
    typedef StrandPairSequence<> strand_pair_sequence_type;
    typedef std::map<duplex_type,std::string>    equivalence_map_type;
    typedef ReferenceSet<Duplex<> >                    duplex_ref_type;
    typedef ReferenceSet<NeighbourSequence<> >         neighbours_ref_type;
    typedef ReferenceSet<StrandPairSequence<> >        strandpair_ref_type;
    typedef std::map<duplex_type,int>             trimers_map_type;
    typedef std::valarray<value_type>                  vector_type;

    sequence_type sequence;
#if __GNUC__ > 5
    constexpr
#endif
    static const value_type Concentration_factor=1e-6; ///< All species concentrations are in milimol
#if __GNUC__ > 5
    constexpr
#endif
    static const value_type R=1.987;   ///< Gas constant, cal/(K mol)
    value_type Ref_concentration;     ///< Reference species concentration
    value_type alpha;                 ///< Determines if sequence is self-complementary
    value_type species_concentration; ///< Oligomer species concentration in \mu M
    value_type salt_concentration;    ///< Salt concentration in mM
    data_type  temperature;           ///< Sequence temperature to species concentration
    data_type  enthalpy;              ///< Sequence Enthalpy as given in papers
    data_type  entropy;               ///< Sequence Entropy as given in papers
    int length;                       ///< Sequence length
    int prediction_method;            ///< Flag for selecting prediction methods
    value_type thermal_equivalence;   ///< Thermal equivalence
    value_type thermal_index;         ///< Thermal index sqrt(thermal_equivalence)
    value_type partition_function;    ///< Zy
    value_type helmholtz_energy;      ///< Fy
    vector_type average_y;            ///< vector containing <y_i>
    bool Periodic;                    ///< Consider sequence as periodic, this will generate an additional NN pair connecting the first and the last bp

    /// Sequence information
    duplex_type            Exact_duplex;        ///< Contains the sequence in Duplex format, it is exact because we are not applying symmetry reductions
    duplex_type            BP_reduced_duplex;      ///< Contains the base-pair symmetry reduced sequence in Duplex format
    duplex_type            Symm_reduced_duplex;    ///< Contains the full-sequence symmetry reduced sequence in Duplex format
    neighboursequence_type Exact_neighbours;    ///< Contains the non-reduced sequence in Nearest neighbour format
    neighboursequence_type Reduced_neighbours;  ///< Contains the NN symmetry reduced sequence in Nearest neighbour format
    strand_pair_sequence_type Exact_strandpairs; ///< Contains the non-symmetry reduced sequence for strand pairs
    strand_pair_sequence_type Reduced_strandpairs;///< Contains the symmetry reduced sequence for strand pairs
    neighboursequence_type NPExact_neighbours;    ///< non-periodic
    neighboursequence_type NPReduced_neighbours;  ///< non-periodic
    strand_pair_sequence_type NPExact_strandpairs; ///< non-periodic
    strand_pair_sequence_type NPReduced_strandpairs;///< non-periodic
    equivalence_map_type*  pEquivalenceMap;     ///< Pointer to Dataset equivalence map
    duplex_deque_type      Trimers_list;
    trimers_map_type       Trimers_map;
    duplex_ref_type        Basepair_set;         ///< The set of different base pairs
    neighbours_ref_type    Neighbours_set;       ///< The set of different base pair neighbours
    neighbours_ref_type    Exact_neighbours_set; ///< The set of different base pair neighbours
    strandpair_ref_type    StrandPair_set;       ///< The set of different strand pair neighbours
    neighbours_ref_type    NPNeighbours_set;     ///< The set of different base pair neighbours, non-periodic
    strandpair_ref_type    NPStrandPair_set;     ///< The set of different strand pair neighbours, non-periodic
    /// Sequence analysis
    bool                Has_CG;                 ///< True if sequence has at least one CG
    bool                Is_AT_only;             ///< True if sequence is AT only (may also be AU) 
    bool                Self_complementary;     ///< True if sequence is self-complementary
    base_pair_type      Terminal5_bp;           ///< Stores the 5' base pair
    base_pair_type      Terminal3_bp;           ///< Stores the 3' base pair
    base_pair_type      Terminal5_bp_reduced;   ///< Stores the symmetry reduced 5' base pair
    base_pair_type      Terminal3_bp_reduced;   ///< Stores the symmetry reduced 3' base pair
    int                 Terminal_5TA3;          ///< Number of 5'TA3' terminals
    int                 Terminal_AU;             ///< Number of AU or UA terminals
    int                 Terminal_AT;             ///< Number of AT or TA terminals
    int                 Terminal_CG;             ///< Number of CG or GC terminals
    int                 Terminal_dTrA;           ///< Number of dTrA or rAdT terminals
    int                 Terminal_dArU;           ///< Number of dArU or rUdA terminals
    int                 Terminal_dCrG;           ///< Number of dTrA or rAdT terminals
    int                 Terminal_dGrC;           ///< Number of dArT or rTdA terminals
 
    SequenceInfo(void): sequence(), alpha(), species_concentration(), 
                        salt_concentration(), 
                        temperature("temperature"), enthalpy("enthalpy"), entropy("entropy"),
                        prediction_method(), thermal_equivalence(), thermal_index(), partition_function(), helmholtz_energy(),
                        Periodic(_Periodic),
                        pEquivalenceMap(),
                        Has_CG(false), Is_AT_only(false),
                        Self_complementary(false), Terminal5_bp(), Terminal3_bp(), Terminal_5TA3(),
                        Terminal_dTrA(), Terminal_dArU(), Terminal_dCrG(), Terminal_dGrC() {}

    inline void print_head(std::ostream& out)
      {
      out << "Main/Complementary " 
          << "alpha "
          << "salt_concentration "
          << "species_concentration ";
      temperature.print_name(out);
      enthalpy.print_name(out);
      entropy.print_name(out);
      out << "prediction_method " 
          << "thermal_equivalence "
          << "thermal_index "
          << "partition_function "
          << "helmholtz_energy "
          << std::endl;
      }

    void print(std::ostream& out, bool analysis=false)
      {
      if (analysis) out << "Dataset entry: ";
      out << sequence.first << "/" << sequence.second 
          << " " << alpha
          << " " << salt_concentration
          << " " << species_concentration
          << " " << temperature  
          << " " << enthalpy 
          << " " << entropy 
          << " " << prediction_method 
          << " " << thermal_equivalence
          << " " << thermal_index
          << " " << partition_function
          << " " << helmholtz_energy
          << std::endl;
      if (analysis)
        {
        out << "* General characteristics " << std::endl;
        out << "Length: " << Exact_duplex.size() << " "
            << "Has_CG: " << Has_CG << " "
            << "Is_AT_only: " << Is_AT_only << " "
            << "Self_complementary: " << Self_complementary << " " << std::endl;
        out << "* Terminal characteristics " << std::endl;
        out << "Terminal base pairs at 5: " << Terminal5_bp << std::endl;
        out << "Terminal base pairs at 3: " << Terminal3_bp << std::endl;
        out << "Terminal base pairs (BP symmetry reduced) at 5: " << Terminal5_bp_reduced << std::endl;
        out << "Terminal base pairs (BP symmetry reduced) at 3: " << Terminal3_bp_reduced << std::endl;
        out << "Terminal_5TA3: " << Terminal_5TA3 << " "
            << "Terminal_AU: " << Terminal_AU << " "
            << "Terminal_AT: " << Terminal_AT << " "
            << "Terminal_CG: " << Terminal_CG << std::endl;
        out << "Terminal_dArU: " << Terminal_dArU << " "
            << "Terminal_dTrA: " << Terminal_dTrA << " "
            << "Terminal_dCrG: " << Terminal_dCrG << " "
            << "Terminal_dGrC: " << Terminal_dGrC << std::endl;
        out << "Original_sequence  : "  << (std::string)Exact_duplex  << std::endl;
        out << "Symmetry_reduced   : "  << (std::string)Symm_reduced_duplex   << std::endl;
        out << "* Sequence decomposition: base pair list" << std::endl;
        out << "Exact_duplex       : "  << Exact_duplex  << " (" << Exact_duplex.size() << ")" << std::endl;
        out << "BP_reduced_duplex  : "  << BP_reduced_duplex  << " (" << BP_reduced_duplex.size() << ")"   << std::endl;
        out << "* Sequence decomposition: nearest-neighbour list" << std::endl;
        out << "Exact_neighbours   : "   << Exact_neighbours  << " (" << Exact_neighbours.size() << ")"   <<  std::endl;
        out << "Reduced_neighbours : " << Reduced_neighbours  << " (" << Reduced_neighbours.size() << ")"   <<  std::endl;
        out << "* Sequence decomposition: intra-strand neighbours list" << std::endl;
        out << "Exact_strandpairs  : " << Exact_strandpairs   << " (" << Exact_strandpairs.size() << ")"   <<  std::endl;
        out << "Reduced_strandpairs: " << Reduced_strandpairs << " (" << Reduced_strandpairs.size() << ")"   <<  std::endl;
        out << "* Sequence elements (symmetry reduced, strand-pairs reduced to 5'->3' direction)" << std::endl;
        out << "Basepair_set       : (" << Basepair_set.size()   << ") " << Basepair_set;
        out << "Neighbours_set     : (" << Neighbours_set.size() << ") " << Neighbours_set;
        out << "Exact_neighbours_set     : (" << Exact_neighbours_set.size() << ") " << Exact_neighbours_set;
        out << "StrandPair_set     : (" << StrandPair_set.size() << ") " << StrandPair_set ;
        typename duplex_deque_type::iterator tri;
        out << "Trimers_list       : (" << Trimers_list.size() << ") ";
        for (tri=Trimers_list.begin(); tri != Trimers_list.end(); tri++)
          {
          out << (std::string)(*tri) << " ";
          }
        typename trimers_map_type::iterator trim;
        out << std::endl;
        out << "Trimers_set        : (" << Trimers_map.size() << ") ";
        for (trim=Trimers_map.begin(); trim != Trimers_map.end(); trim++)
          {
          out << (std::string)(trim->first) << "=" << trim->second << " ";
          }
        out << std::endl << std::endl;
      
        }
      }
      
    inline void apply_equivalence_map(void)
      {
      //Generates a list of trimers which includes both ends, 
      //for example CGGCUG/GUCGGC will result in UG_/GC_ CGG/GUC GCU/CGG GCU/CGG CGG/GUC UG_/GC_
      //note: this requires underscore (_) to be included in NucleotideDictionary
      if (pEquivalenceMap != NULL)
        {
        int duplex_length=static_cast<int>(Exact_duplex.size());
        std::deque<typename duplex_type::base_pair_type::version_type> version_list(Exact_duplex.size());
        for(int di=-1; di < (duplex_length-1); di++)
          {
          std::string termstr("__");
          if (Exact_duplex.is_hybrid()) termstr=Exact_duplex[0].first.sugar_string() + "_" + Exact_duplex[0].second.sugar_string() + "_";
          BasePair<>  terminal(termstr);
          duplex_type tmp;
          if (di < 0) tmp.push_back(terminal);
          else        tmp.push_back(Exact_duplex[di]);
          tmp.push_back(Exact_duplex[di+1]);
          if (di >= (duplex_length-2)) tmp.push_back(terminal);
          else        tmp.push_back(Exact_duplex[di+2]);
          tmp.reduce_to_smallest_duplex_symmetry();
          equivalence_map_type::iterator eq=pEquivalenceMap->find(tmp);
          if (eq != pEquivalenceMap->end())
            {
            BasePair<> rep(eq->second);
            version_list[di+1]=rep.Version;
            }
          Trimers_list.push_back(tmp);
          Trimers_map[tmp]++;
          }
        for(int di=0; di < duplex_length; di++)
          {
          if (version_list[di].length() > 0)
            {
            Exact_duplex[di].Version=version_list[di];
            BP_reduced_duplex[di].Version=version_list[di];
            Symm_reduced_duplex[di].Version=version_list[di];
            }
          }
        }
      }
      
      
    // This is a crucial function, here we take the two sequence strings, one for main strand
    // one for complementary strand, and work out all equivalences, base pairs, nearest-neighbours,
    // determine if the are AT or CG terminals etc
    inline void insert_sequence(const std::string& seq, const std::string& comp=std::string())
      {
      Exact_duplex.clear();
      BP_reduced_duplex.clear();
      Symm_reduced_duplex.clear();
     if(comp != std::string())
        {
        Exact_duplex=Duplex<>(seq+"/"+comp,BasePair<>::do_not_simplify_symmetry);
        BP_reduced_duplex=Duplex<>(seq+"/"+comp,BasePair<>::simplify_symmetry);
        Symm_reduced_duplex=Duplex<>(seq+"/"+comp,BasePair<>::do_not_simplify_symmetry);
        Symm_reduced_duplex.reduce_to_smallest_duplex_symmetry();
        }
      else 
        {
        Exact_duplex=Duplex<>(seq,BasePair<>::do_not_simplify_symmetry);
        BP_reduced_duplex=Duplex<>(seq,BasePair<>::simplify_symmetry);
        Symm_reduced_duplex=Duplex<>(seq,BasePair<>::do_not_simplify_symmetry);
        Symm_reduced_duplex.reduce_to_smallest_duplex_symmetry();
        }
        
      apply_equivalence_map(); //this is the map where the equivalence rules read from .par files are applied

      Reduced_neighbours.clear();
      Reduced_neighbours.Periodic=Periodic; // This avoids that the first and last base-pair form a nearest-neighbour pair

      Reduced_neighbours=Exact_duplex;
      Reduced_neighbours.reduce_to_smallest_symmetry();
     
      Exact_neighbours.clear();
      Exact_neighbours.Periodic=Periodic; // This avoids that the first and last base-pair form a nearest-neighbour pair
      Exact_neighbours=Exact_duplex; //here we do not call reduce_to_smallest_symmetry()
      
      Basepair_set.clear();
      Basepair_set.Symmetry_action=BasePair<>::simplify_symmetry;
      Basepair_set.analyse(BP_reduced_duplex);

      Neighbours_set.clear();
      Neighbours_set.analyse(Reduced_neighbours);
      Exact_neighbours_set.clear();
      Exact_neighbours_set.analyse(Exact_neighbours);

      Exact_strandpairs.clear();
      Exact_strandpairs.Symmetry_action=StrandPair<>::do_not_simplify_symmetry;
      Exact_strandpairs = Exact_neighbours;
      
      Reduced_strandpairs.clear();
      Reduced_strandpairs.Symmetry_action=StrandPair<>::simplify_symmetry;
      Reduced_strandpairs = Exact_neighbours;
      
      StrandPair_set.clear();
      StrandPair_set.analyse(Reduced_strandpairs);
      
      if (Periodic)//when periodic we also generate the non-periodic info
        {
        NPReduced_neighbours.clear();
        NPReduced_neighbours.Periodic=false; // This avoids that the first and last base-pair form a nearest-neighbour pair
        NPReduced_neighbours=Exact_duplex;
        NPReduced_neighbours.reduce_to_smallest_symmetry();
        NPExact_neighbours.clear();
        NPExact_neighbours.Periodic=false;
        NPExact_neighbours=Exact_duplex; //here we do not call reduce_to_smallest_symmetry()
        NPNeighbours_set.clear();
        NPNeighbours_set.analyse(NPReduced_neighbours);
        NPExact_strandpairs.clear();
        NPExact_strandpairs.Symmetry_action=StrandPair<>::do_not_simplify_symmetry;
        NPExact_strandpairs = NPExact_neighbours;
        NPReduced_strandpairs.clear();
        NPReduced_strandpairs.Symmetry_action=StrandPair<>::simplify_symmetry;
        NPReduced_strandpairs = NPExact_neighbours;
        NPStrandPair_set.clear();
        NPStrandPair_set.analyse(NPReduced_strandpairs);
        }
        
      Self_complementary=Exact_duplex.self_complementary();
            
      Terminal5_bp = *Exact_duplex.begin();
      Terminal3_bp = *Exact_duplex.rbegin();
      Terminal5_bp_reduced = *BP_reduced_duplex.begin();
      Terminal3_bp_reduced = *BP_reduced_duplex.rbegin();
            
      BasePair<> AT("AT",BasePair<>::simplify_symmetry),
                 AU("AU",BasePair<>::simplify_symmetry),
                 TA("TA",BasePair<>::simplify_symmetry),//internally represented as AT
                 UA("UA",BasePair<>::simplify_symmetry),//internally represented as AU
                 CG("CG",BasePair<>::simplify_symmetry),
                 GC("GC",BasePair<>::simplify_symmetry),//internally represented as CG
                 dTrA("dTrA",BasePair<>::simplify_symmetry),
                 dArU("dArU",BasePair<>::simplify_symmetry),
                 dCrG("dCrG",BasePair<>::simplify_symmetry),
                 dGrC("dGrC",BasePair<>::simplify_symmetry),
                 rAdT("rAdT",BasePair<>::simplify_symmetry),
                 rUdA("rUdA",BasePair<>::simplify_symmetry),
                 rGdC("rGdC",BasePair<>::simplify_symmetry),
                 rCdG("rCdG",BasePair<>::simplify_symmetry);
                 
      Terminal_5TA3=0;
      //This accounts for a 5'T.A3' at the beginning
      if (Terminal5_bp == TA) Terminal_5TA3++;
      //This accounts for a 3'A.T5' at the end
      if (Terminal3_bp  == AT) Terminal_5TA3++;

      Terminal_AU=0;
      //This accounts for a AU or UA at the beginning
      if (Terminal5_bp == AU || Terminal5_bp == UA) Terminal_AU++;
      if (Terminal3_bp == AU || Terminal3_bp == UA) Terminal_AU++;

      Terminal_AT=0;
      //This accounts for a AT or TA at the beginning
      if (Terminal5_bp == AT || Terminal5_bp == TA) Terminal_AT++;
      if (Terminal3_bp == AT || Terminal3_bp == TA) Terminal_AT++;

      Terminal_CG=0;
      //This accounts for a CG or GC at the beginning
      if (Terminal5_bp == CG || Terminal5_bp == GC) Terminal_CG++;
      if (Terminal3_bp == CG || Terminal3_bp == GC) Terminal_CG++;

      Terminal_dArU=0;
      if (Terminal5_bp == dArU || Terminal5_bp == rUdA ) Terminal_dArU++;
      if (Terminal3_bp == dArU || Terminal3_bp == rUdA) Terminal_dArU++;

      Terminal_dTrA=0;
      if (Terminal5_bp == dTrA || Terminal5_bp == rAdT) Terminal_dTrA++;
      if (Terminal3_bp == dTrA || Terminal3_bp == rAdT) Terminal_dTrA++;

      Terminal_dCrG=0;
      if (Terminal5_bp == dCrG || Terminal5_bp == rGdC) Terminal_dCrG++;
      if (Terminal3_bp == dCrG || Terminal3_bp == rGdC) Terminal_dCrG++;

      Terminal_dGrC=0;
      if (Terminal5_bp == dGrC || Terminal5_bp == rCdG) Terminal_dGrC++;
      if (Terminal3_bp == dGrC || Terminal3_bp == rCdG) Terminal_dGrC++;

      Has_CG=false;
      Is_AT_only=true;
      typename duplex_type::iterator it;
      for (it=BP_reduced_duplex.begin(); it != BP_reduced_duplex.end(); ++it) //Attention: do not use BP_reduced_duplex for NN analysis
        {
        if (*it == CG) Has_CG=true;
        if ((not (*it == AT)) and (not (*it == AU)) ) Is_AT_only=false; 
        }
      }
      

    /// \brief Checks for selfcomplementarity and sets alpha value
    ///
    /// \returns 1 if self-complementary, 4 otherwise
    inline value_type determine_alpha(void)
      {
      NucleotideSequence<> s(sequence.first.length()),c(sequence.second.length());
      s << sequence.first; c << sequence.second; c.reverse();
      if ((std::string)s == (std::string)c) alpha=1.0;
      else                                  alpha=4.0;
      return alpha;
      }

    /// \brief Adjust temperature to current reference concentration
    ///
    /// \attention This temperature adjustment is too approximate and should not be used.
    inline void adjust_temperature(void)
      {
      if (Ref_concentration != species_concentration)
        {
        // Note: it is correct to use the same alpha for conc1 and conc2
        temperature.adjusted=temperature.measured*log(species_concentration*Concentration_factor/alpha)/log(Ref_concentration*Concentration_factor/alpha);
        }
      else temperature.adjusted=temperature.measured;
      }

    // Calculates new Tm for given Ct, but does not update any sequence info
    // ct in \micro M (1e-6 M)
    inline value_type calculate_temperature_from_gibbs(value_type &ct)
      {
      return enthalpy.measured*1000.0/(entropy.measured+R*log(ct*Concentration_factor/alpha))-273.15;
      }

    // Updates Tm for default Ct, updates sequence info
    inline void set_temperature_from_gibbs(void)
      {
      temperature.measured=calculate_temperature_from_gibbs(species_concentration);
      temperature.adjusted=temperature.measured;
      }

    
  };
};
#endif
